# OPERATION Filter
# LANGUAGE_SYNTAX 
# "Filter" Syntax:
#   stream.filter( name = "name", \t\t\t\t => optional
#                  emits = "stream", \t\t\t\t => optional
#                  prepare = prepare_function, \t => optional if no initialization needed
#                  keep = keep_function )
#   - A "filter" may only emit a single stream.
#   - The "prepare_function" must take in a single argument (the "controller") and should return nothing. This is where any setup is done to prepare for tuple processing in the "keep_function".
#   - The "keep_function" can be a full function or a lambda function. It must take in a single argument (the tuple), and return boolean "True" or "False". Tuples will pass through if "keep_function" returns "True"."""
# EXAMPLE 
# # Checks for tuples whose "url" field begin with "https"
# def keep_https(tuple):
#   return tuple["url"].starts_with("https")
#  
# stream = stream.filter(
#   name = "https_filter",
#   keep = keep_keep_https
# )
# 
import traceback
from counter import Counter
from controller import Controller, ParentDeadException
from helper import Helper

class Filter:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, emits, prepare, keep, parallelism):
    self._name = name if name else "filter_"+Counter.get()
    self._type = "filter"

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._prepare = prepare
    self._keep = keep
    self._parallelism = parallelism

    Helper.check_name("filter", self._name, self._app._names)
    Helper.check_filter(self)
    Helper.check_emits("filter", self._emits, self._app._streams)

  def run_operation(self):
    host = self._app._options.get("host", None)
    port = self._app._options.get("port", None)
    controller = Controller(host, port, self._emits)

    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" in d and d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        else:
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          if self._keep(tup):
            controller.emit(tup)
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in filter: "+traceback.format_exc())
      controller.done()



# OPERATION Source
# LANGUAGE_SYNTAX 
#   Sourcing from a relation:
#     app.source( matches = "SQL query" or [SXP queries] )
# 
#   Custom source:
#     app.source( name = "name", \t\t\t\t\t => optional
#                 emits = ["stream_1", "stream_2", ...], \t\t => optional for single output stream
#                 end_cycle_policy = "null_emit" OR "explicit", \t => default "null_emit"
#                 begin_cycle = begin_cycle_function, \t\t => optional if no initialization needed
#                 next_tuple = next_tuple_function )
# 
#   - The "end_cycle_policy" is used to specify when a cycle should end. Two options are available:
#       * :null_emit - end the cycle when a field contains "nil" or when nothing is emitted from the "next_tuple" block.
#       * :explicit - the end of a cycle is explicitly declared in the "next_tuple" block. This is done by including the "end_cycle" keyword in the "next_tuple" function, e.g. end_cycle() if queue.empty.
#   - The "begin_cycle_function" and "next_tuple_function" can be full functions or lambda functions. 
#       * "begin_cycle_function" must take in a single argument (the "controller") and should return nothing. This is where any setup is done to initialize the content and quantity of tuples emitted by "next_tuple_function".
#       * "next_tuple_function" must take in 2 arguments (the "controller" and the "tuple"), and should return nothing. This is where the tuples are actually emitted."""
# EXAMPLE 
# # Source from the "homepages" relation
# stream = app.source(name="homepages")
#  
#  
#  
# # Custom Source
# source_pages = ["http://www.zillabyte.com", "http://docs.zillabyte.com", "http://www.zillabyte.com/about", "http://blog.zillabyte.com"]
#  
# def nt(controller):
#   if(len(source_pages) > 0):
#     controller.emit({"url":source_pages.pop(0)})
#   else:
#     controller.end_cycle()
#  
# app = zillabyte.app(name="python_app")
#  
# stream = app.source(name="python_source", next_tuple=nt, end_cycle_policy="explicit")
import traceback
from counter import Counter
from controller import Controller, ParentDeadException, UserEndCycleException
from helper import Helper

class Source:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, matches, options, emits, end_cycle_policy, prepare, begin_cycle, next_tuple):
    self._name = name if name else "source_"+Counter.get()
    self._type = "source"
    self._options = options
    self._matches = None
    self._relation = None
    self._end_cycle_policy = None

    if(matches != None):
      Helper.check_matches(matches)
      if isinstance(matches, basestring):
        if not isinstance(options, dict):
          options = {}
        self._relation = {"query": matches, "options": options}
      elif isinstance(matches, list):
        self._matches = matches
    else:
      self._end_cycle_policy = end_cycle_policy

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._prepare = prepare
    self._begin_cycle = begin_cycle
    self._next_tuple = next_tuple

    Helper.check_name("source", self._name, self._app._names)
    Helper.check_source(self)
    Helper.check_emits("source", self._emits, self._app._streams)

  def run_operation(self):
    host = self._app._options.get("host", None)
    port = self._app._options.get("port", None)
    controller = Controller(host, port, self._emits)

    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" not in d:
          controller.log("Not a command")
        elif d["command"] == "prepare": 
          if self._prepare != None:
            self._prepare(controller)
        elif d["command"] == "begin_cycle":
          if(self._begin_cycle != None):
            self._begin_cycle(controller)
        elif d["command"] == "next":
            self._next_tuple(controller)
        elif d["command"] == "ack":
          controller.ack(d.get("id",None))
        elif d["command"] == "fail":
          controller.fail(d.get("id",None))
        else:
          controller.log("Invalid command")
      except UserEndCycleException, e:
        pass
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in source: "+traceback.format_exc())
      controller.done()



from counter import Counter
from helper import Helper

class ComponentInput:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, fields):
    self._name = name
    self._type = "input"

    self._emits = ["stream_"+Counter.get()]
    self._fields = fields

    Helper.check_name("inputs", self._name, self._app._names)
    Helper.check_component_source(self)

  def run_operation(self):
    print "Component sources cannot be run!"



from counter import Counter
from helper import Helper

class InjectedComponent:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, idn, input_stream_1, additional_inputs, outputs, output_format):
    self._name = name if name else "component_"+Counter.get()
    self._id = str(idn)
    self._type = "component"
    self._consumes = [input_stream_1]

    if isinstance(additional_inputs, list):
      self._consumes += additional_inputs
    elif additional_inputs != None:
      self._consumes.append(additional_inputs)

    if outputs == None:
      outputs = ["stream_"+Counter.get()]
    elif isinstance(outputs, basestring):
      outputs = [outputs]
    self._emits = outputs

    self._output_format = output_format

    Helper.check_name("component", self._name, self._app._names)
    Helper.check_call_component(self)
    Helper.check_emits("component", self._emits, self._app._streams)

  def run_operation(self):
    print "Embedded components are only instantiated on our cluster, sorry!"



# OPERATION Join
# LANGUAGE_SYNTAX 
# lhs_stream.join_with( name = "name", \t\t\t\t\t\t\t\t => optional
#                       stream = rhs_stream_object,
#                       fields = "join_field" OR ["lhs_join_field", "rhs_join_field"],
#                       emits = "stream", \t\t\t\t\t\t\t => optional
#                       options = {options} )
# - A "join" may only emit a single stream.
# - The following options are supported:
#     * "type" \t -- specifies the join type. The default is "left".
# EXAMPLE 
#  
# word_stream = employee_stream.join_with(
#   name = "join",
#   stream = budget_stream,
#   fields = ["occupation", "number"],
#   options = {"type" : "left"})
from counter import Counter
from helper import Helper
class Join:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, lhs_stream, rhs_stream, emits, fields, options):
    self._name = name if name else "join_"+Counter.get()
    self._type = "join"
    self._lhs_stream = lhs_stream._name
    self._rhs_stream = rhs_stream._name

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    if fields:
      if isinstance(fields, list) and len(fields) == 2:
        self._lhs_fields = fields[0]
        self._rhs_fields = fields[1]
      elif isinstance(fields, basestring):
        self._lhs_fields = fields
        self._rhs_fields = fields

    self._join_type = "inner"
    if options:
      if options["type"]:
        self._join_type = options["type"]

    Helper.check_name("join", self._name, self._app._names)
    Helper.check_join(self)
    Helper.check_emits("join", self._emits, self._app._streams)

  def run_operation(self):
    print "Joins can only be instantiated on our cluster, sorry!"



from controller import *
from stream import Stream
from operation_handler import OperationHandler
from helper import Helper
from base import Base
from version import VERSION

class App(Base):
  def __init__(self, name=None):
    Base.__init__(self, name)
    Helper.check_name("app", self._name, {})
    self._options = Helper.opt_parser()

    if(self._options["command"] == "info"):
      sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      sock.connect((self._options["host"], int(self._options["port"])))
      self._socket = sock.makefile()
      Helper.write_hash_to_file({"language":"python", "name":self._name, "flow_type":"app", "multilang_version":VERSION}, self._socket)

  def source(self, name=None, matches=None, options={}, emits=None, end_cycle_policy="null_emit", prepare=None, begin_cycle=None, next_tuple=None):
    op = OperationHandler(self, Stream)
    streams = op.build_multilang_operation("source", name, matches, options, emits, end_cycle_policy, begin_cycle, next_tuple)\
                       .add_operation_properties_to_info("name", "type")\
                       .add_optional_operation_properties_to_info("relation", "matches", "end_cycle_policy")\
                       .handle_operation()\
                       .get_output_streams()
    return streams



class Tuple:
  def __init__(self, tup):
    self.values = tup["tuple"]
    self.meta = tup["meta"]
    self.column_aliases = tup.get("column_aliases", [])
    self.alias_hash = None

  def __getitem__(self, k):
    self.maybe_build_alias_hash()
    concrete_name_to_get = self.alias_hash.get(k, None)
    concrete_name_to_get = concrete_name_to_get if concrete_name_to_get else k
    if concrete_name_to_get in self.meta:
      return self.meta[concrete_name_to_get]
    elif concrete_name_to_get in self.values:
      return self.values[concrete_name_to_get]
    else:
      return None

  def maybe_build_alias_hash(self):
    if not self.alias_hash:
      self.alias_hash = {}
      for col in self.column_aliases:
#        rel_name = col["relation_name"]
        concrete_name = col["concrete_name"]

        # Note that this won't work with multiple tables since concrete_name will be repeated amongst them
        self.alias_hash[col["alias"]] = concrete_name
    return



from counter import Counter

class JvmOperation:
  def __init__(self, op_type):
    self._type = op_type

  def build_node(self, *args):
    self._name = self._type+"_"+Counter.get()
    self._emits = ["stream_"+Counter.get()]
    self._args = args

  def run_operation(self):
    print self._type+" operation can only be instantiated on our cluster, sorry!"



from helper import Helper
from operation_handler import OperationHandler

class Stream:
  def __init__(self, name, app, previous_node_name):
    self._name = name
    self._app = app
    self._previous_node_name = previous_node_name

  def each(self, name=None, emits=None, output_format="replace", prepare=None, execute=None, parallelism=None):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_multilang_operation("each", name, emits, output_format, prepare, execute, parallelism)\
                 .add_operation_properties_to_info("name", "type", "output_format")\
                 .add_optional_operation_properties_to_info("parallelism")\
                 .create_arc_info_from_stream(self)\
                 .handle_operation()\
                 .get_output_streams()
    return streams

  def filter(self, name=None, emits=None, prepare=None, keep=None, parallelism=None):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_multilang_operation("filter", name, emits, prepare, keep, parallelism)\
                .add_operation_properties_to_info("name", "type")\
                .add_optional_operation_properties_to_info("parallelism")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def group_by(self, name=None, emits=None, fields=None, prepare=None, begin_group=None, aggregate=None, end_group=None, parallelism=None):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_multilang_operation("group_by", name, emits, fields, begin_group, aggregate, end_group, parallelism)\
                .add_operation_properties_to_info("name", "type", "group_by")\
                .add_optional_operation_properties_to_info("parallelism")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def join_with(self, name=None, stream=None, emits=None, fields=None, options=None):
    lhs_stream = self
    rhs_stream = stream
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_multilang_operation("join", name, self, stream, emits, fields, options)\
                .add_operation_properties_to_info("name", "type", "lhs_stream", "rhs_stream", "lhs_fields", "rhs_fields", "join_type")\
                .create_arc_info_from_stream(lhs_stream, "left")\
                .create_arc_info_from_stream(rhs_stream, "right")\
                .handle_operation()\
                .get_output_streams()
    return streams

  def call_component(self, name=None, component_id=None, additional_inputs=None, outputs=None, output_format="replace"):
    op = OperationHandler(self._app, self.__class__)
    op = op.build_multilang_operation("component", name, component_id, self, additional_inputs, outputs, output_format)\
           .add_operation_properties_to_info("name", "type", "id", "output_format")
    node = op.node()
    for stream in node._consumes:
      op = op.create_arc_info_from_stream(stream)
    streams = op.handle_operation()\
                .get_output_streams()
    return streams

  def loop_back(self, operation):
    Helper.check_loop_back(self, operation, self._app._nodes)
    Helper.write_arc_to_file({"name" : self._name, "origin" : self._previous_node_name, "dest" : operation, "loop_back" : 1}, self._app._socket)
    return self

  def sink(self, name=None, columns=None, relation=None, scope=None):
    op = OperationHandler(self._app, self.__class__)
    op.build_multilang_operation("sink", name, columns, relation, scope)\
      .add_operation_properties_to_info("name", "type", "columns", "relation", "scope")\
      .create_arc_info_from_stream(self)\
      .handle_operation()

  def unique(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("unique", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("group_fields")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def count(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("count", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("group_fields")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def remove(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("remove", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("remove")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def retain(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("retain", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("retain")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def clump(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("clump", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("clump")\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams

  def rename(self, *args):
    op = OperationHandler(self._app, self.__class__)
    streams = op.build_jvm_operation("rename", *args)\
                .add_operation_properties_to_info("name", "type")\
                .add_input_args_to_info_as("rename", 0)\
                .create_arc_info_from_stream(self)\
                .handle_operation()\
                .get_output_streams()
    return streams



from controller import *
from counter import Counter
from helper import Helper
from component_stream import ComponentStream
from operation_handler import OperationHandler
from base import Base
from version import VERSION

class Component(Base):
  def __init__(self, name=None):
    Base.__init__(self, name)
    Helper.check_name("component", self._name, {})
    self._options = Helper.opt_parser()

    if(self._options["command"] == "info"):
      sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      sock.connect((self._options["host"], int(self._options["port"])))
      self._socket = sock.makefile()
      Helper.write_hash_to_file({"language":"python", "name":self._name, "flow_type":"component", "multilang_version":VERSION}, self._socket)

  def inputs(self, name=None, fields=None):
    op = OperationHandler(self, ComponentStream)
    streams = op.build_multilang_operation("component_input", name, fields)\
                .add_operation_properties_to_info("name", "type", "fields")\
                .handle_operation()\
                .get_output_streams()
    return streams



from helper import Helper

class ComponentOutput:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, columns, relation, scope):
    self._name = name
    self._type = "output"
    self._columns = columns

    if relation != None:
      self._relation = relation
    else:
      self._relation = name
    self._scope = scope

    Helper.check_sink(self._name, self._columns, self._app._nodes, "outputs")

  def run_operation(self):
    print "Component sinks cannot be run!"
